#!/usr/bin/perl

eval 'exec /usr/bin/perl  -S $0 ${1+"$@"}'
    if 0; # not running under some shell

use strict;
use warnings;
use lib qw/lib/;
use Urb::Dom;
use Urb::Dom::Mutate;
use YAML;

my $DEBUG = $ENV{DEBUG};

my $path_yaml_a = shift || die "Usage $0 some_dom_object.yml another_dom_object.yml";
my $path_yaml_b = shift || die;

my $dom_a = new Urb::Dom;
my $dom_b = new Urb::Dom;
$dom_a->Deserialise (YAML::LoadFile ($path_yaml_a));
$dom_b->Deserialise (YAML::LoadFile ($path_yaml_b));

my $mutator = new Urb::Dom::Mutate;

# 80% recombination 20% mutation
if (int rand (5))
{
    # pick a random quad from A
    my $victim_a = victim ($dom_a);

    # pick a random quad from B from the nearest third by area
    my $victim_b = victim_byarea ($dom_b, $victim_a->Area);
    print "crossover victims: ". $victim_a->Id .' '. $victim_b->Id ."\n" if $DEBUG;

    # no point combining a wide quad with a narrow quad, rotate to fix
    my $aspect_a = ($victim_a->Length (0) + $victim_a->Length (2)) /
                   ($victim_a->Length (1) + $victim_a->Length (3));
    my $aspect_b = ($victim_b->Length (0) + $victim_b->Length (2)) /
                   ($victim_b->Length (1) + $victim_b->Length (3));
    if (($aspect_a < 1 and $aspect_b > 1) or ($aspect_a > 1 and $aspect_b < 1))
    {
        int rand (2) ? $victim_b->Rotate : $victim_b->Unrotate;
        print "  rotating to fit\n" if $DEBUG;
    }

    # exchange victims
    $victim_a->Crossover ($victim_b);

    # do one mutation in A
    $dom_a = $mutator->apply ($dom_a);
}
else
{
    print "mutating\n" if $DEBUG;
    # do one mutation in A
    $dom_a = $mutator->apply ($dom_a);

    # do two mutations in B
    $dom_b = $mutator->apply ($dom_b);
    $dom_b = $mutator->apply ($dom_b);
}

$dom_a->Merge_Divided;
$dom_a->Split_Undivided_Up;
$dom_a->Highest->Split_Undivided_Down;
$dom_b->Merge_Divided;
$dom_b->Split_Undivided_Up;
$dom_b->Highest->Split_Undivided_Down;

my $hash_a = $dom_a->Hash;
my $hash_b = $dom_b->Hash;

YAML::DumpFile ($hash_a .'.yml', $dom_a->Serialise);
YAML::DumpFile ($hash_b .'.yml', $dom_b->Serialise);

sub victim
{
    my $dom = shift;
    my @all = map {$_->Children} $dom, $dom->Levels_Above;
    my $index = int rand (scalar @all);
    return $all[$index];
}

sub victim_byarea
{
    my $dom = shift;
    my $area = shift;
    my $list = $dom->By_Area ($area);
    my $toget = int (scalar (@{$list}) / 3) + 1;
    my $index = int rand ($toget);
    return $list->[$index]->[1];
}

exit 0;

__END__

=head1 NAME

urb-combine - Crossover two Urb::Dom files

=head1 SYNOPSIS

urb-combine <some_dom_object.yml> <another_dom_object.yml>

=head1 DESCRIPTION

B<urb-combine> takes two L<Urb::Dom> objects that have been serialised as
L<YAML> files, selects a sub-branch from each tree and exchanges them, this is
analogous to genetic crossover in living organisms.  The resulting offspring
are then written in L<YAML> format to disk using filenames derived from a hash
of the serialised content.

Note that currently a mutation is also applied using L<Urb::Dom::Mutate>.

=head1 COPYRIGHT

Copyright (c) 2004-2011 Bruno Postle <bruno@postle.net>.

This file is part of Urb.

Urb is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Urb is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Urb.  If not, see <http://www.gnu.org/licenses/>.

=head1 SEE ALSO

L<Urb>

=cut

