#!/usr/bin/perl

eval 'exec /usr/bin/perl  -S $0 ${1+"$@"}'
    if 0; # not running under some shell

use strict;
use warnings;
use lib qw/lib/;
use Urb::Dom;
use Urb::Polygon;
use YAML;
use Graph;
use Math::Trig;
use 5.010;

my $dom_all = new Urb::Dom;
$dom_all->Deserialise (YAML::LoadFile ($ARGV[0]));

$dom_all->Merge_Divided;
$dom_all->Split_Undivided_Up;
$dom_all->Highest->Split_Undivided_Down;

my $yaml_id = 0;

for my $dom ($dom_all, $dom_all->Levels_Above)
{
    my $boundaries = $dom->Calc_Boundaries;
    my $graph = $dom->Graph;
    my $elevation = 3.0 * scalar ($dom->Levels_Below);
    my $height = 3.0;

    my $graph_envelope = new Graph (directed => 1);
    my $graph_partywall = new Graph (directed => 1);
    my $graph_private = new Graph (directed => 1);
    my $graph_private_covered = new Graph (directed => 1);
    my $graph_covered = new Graph (directed => 1);
    my $graph_public = new Graph (directed => 1);
    my $graph_roofline = new Graph (directed => 1);
    my $graph_parapet = new Graph (directed => 1);

    my $graph_internal = new Graph (undirected => 1);
    my $lookup;

    for my $vertex ($graph->vertices)
    {  
        for my $neighbour ($graph->neighbours ($vertex))
        {
            for my $key_boundary (keys %{$boundaries})
            {
                my $overlap = $boundaries->{$key_boundary}->Overlap ($vertex, $neighbour);
                next if $overlap == 0;
                my @coordinates = $boundaries->{$key_boundary}->Coordinates ($vertex, $neighbour);
                my $node0 = join ':', @{$coordinates[0]};
                my $node1 = join ':', @{$coordinates[1]};

                $lookup->{$node0.'-'.$node1} = [$vertex->Id, $neighbour->Id];

                # zero length links result in infinite loops
                next if $node0 eq $node1;
                if ($vertex->Type !~ /^o/i and $neighbour->Type =~ /^o/i)
                {
                    $graph_envelope->add_edge ($node0, $node1);
                    $graph_roofline->add_edge ($node0, $node1) unless scalar $dom->Levels_Above;
                }
                elsif ($vertex->Type !~ /^o/i)
                {
                    $graph_internal->add_edge ($node0, $node1);
                }
                if ($vertex->Type =~ /^o/i and $neighbour->Type =~ /^o/i and $vertex->Is_Supported and !$neighbour->Is_Supported)
                {
                    $graph_parapet->add_edge ($node0, $node1);
                }
                if ($vertex->Type =~ /^o/i and $neighbour->Type =~ /^o/i and $vertex->Is_Covered and !$neighbour->Is_Covered)
                {
                    $graph_covered->add_edge ($node0, $node1);
                }
            }
        }
        for my $id (0 .. 3)
        {  
            my $external = $vertex->Boundary_Id ($id);
            next unless ($external =~ /^[abcd]$/);
            my $external_type = $vertex->Root->Perimeter->{$external} || '';
            my $node0 = join ':', @{$vertex->Coordinate ($id)};
            my $node1 = join ':', @{$vertex->Coordinate ($id +1)};

            $lookup->{$node0.'-'.$node1} = [$vertex->Id, $external];

            if ($vertex->Type =~ /^o/i)
            {
                $graph_public->add_edge ($node0, $node1) if $external_type ne 'private';
                $graph_private->add_edge ($node0, $node1) if ($external_type eq 'private' and !$vertex->Is_Covered);
                $graph_private_covered->add_edge ($node0, $node1) if ($external_type eq 'private' and $vertex->Is_Covered);
                $graph_covered->add_edge ($node0, $node1) if ($external_type ne 'private' and $vertex->Is_Covered);
                $graph_parapet->add_edge ($node0, $node1) if ($vertex->Is_Supported);
            }
            elsif ($external_type eq 'private')
            {
                $graph_partywall->add_edge ($node0, $node1);
                $graph_roofline->add_edge ($node0, $node1);
            }
            else
            {
                $graph_envelope->add_edge ($node0, $node1);
                $graph_roofline->add_edge ($node0, $node1);
            }
        }
        # ceiling
        unless ($vertex->Is_Outside and !$vertex->Is_Covered)
        {
            my $molior = {type => 'molior-ceiling', elevation => $elevation, closed => 1, height => $height, inner => 0.08, path => [map {$vertex->Coordinate ($_)} 0 .. 3]};
            YAML::DumpFile ("molior-$yaml_id.yml", $molior);
            $yaml_id++;
        }
        unless ($vertex->Is_Outside)
        {
            my $molior = {type => 'molior-floor', elevation => $elevation, closed => 1, height => $height, inner => 0.08, path => [map {$vertex->Coordinate ($_)} 0 .. 3]};
            YAML::DumpFile ("molior-$yaml_id.yml", $molior);
            $yaml_id++;
        }
    }

    # do roof
    my @perimeters;
    my @holes;
    while ($graph_roofline->has_a_cycle)
    {
        my @chain = $graph_roofline->find_a_cycle;
        $graph_roofline->delete_vertices (@chain);

        my $chain = _unpack_chain (@chain);
        my $polygon = new Urb::Polygon (points => [@{$chain}, $chain->[0]]);
        if ($polygon->isClockwise)
        { push @holes, $polygon }
        else
        { push @perimeters, $polygon }
    }

    my @quads;
    for my $perimeter (@perimeters)
    {
        push @quads, $perimeter->_to_quads (@holes);
    }

    my @privates;
    my $id = 0 - $dom->Rotation;
    for my $edge ('a', 'b', 'c', 'd')
    {
        push @privates, [$dom->Coordinate ($id), $dom->Coordinate ($id+1)] if (defined $dom->Perimeter->{$edge} and $dom->Perimeter->{$edge} eq 'private');
        $id++;
    }

    for my $quad (@quads)
    {
        my @points = $quad->points;

        my $gable;
        for my $id (0 .. 3)
        {
            my $midpoint = Urb::Polygon::_scale_2d (1/2, Urb::Polygon::_add_2d ($points[$id],$points[$id+1]));
            my $gableness = 0;
            for my $private_edge (@privates)
            {
                $gableness = 1 if Urb::Polygon::_is_between_2d ($midpoint, @{$private_edge});
            }
            $gable .= $gableness;
        }

        pop @points;
        my $molior = {capbottom => 1, type => 'molior-roof', elevation => $elevation, gradient => 0.66666, closed => 1, gable => $gable,
                      height => $height, outer => 0.25, path => [@points]};

        YAML::DumpFile ("molior-$yaml_id.yml", $molior);
        $yaml_id++;
    }

    # do looped external walls
    while ($graph_envelope->has_a_cycle)
    {
        my $molior = {closed => 1, captop => 1, capbottom => 1, type => 'molior-wall',
            elevation => $elevation, height => $height, inner => 0.08, outer => 0.25,
            path => [], openings => []};

        my @chain = $graph_envelope->find_a_cycle;
        $molior->{path} = _unpack_chain (@chain);

        for my $edge_id (0 .. scalar @{$molior->{path}} -2, -1)
        {
            _insert_opening ($dom, $molior, $lookup, [@chain], $edge_id);
        }

        YAML::DumpFile ("molior-$yaml_id.yml", $molior);
        $yaml_id++;

        unless (scalar $dom->Levels_Above)
        {
	    my $molior_extrusion = {closed => 1, type => 'molior-extrusion',
               elevation => $elevation, height => $height, path => _unpack_chain (@chain),
               profile => [[0.25,0.0], [0.25,-0.45], [0.3,-0.45], [0.3,-0.4], [0.35,-0.4],
                           [0.35,-0.35], [0.5,-0.25], [0.55,-0.25], [0.55,-0.2]]};

            YAML::DumpFile ("molior-$yaml_id.yml", $molior_extrusion);
            $yaml_id++;
        }

        $graph_envelope->delete_vertices (@chain);
    }

    # do non-loop external walls
    for my $source_vertex ($graph_envelope->source_vertices)
    {
        my $molior = {closed => 0, captop => 1, capbottom => 1, type => 'molior-wall',
            elevation => $elevation, height => $height, inner => 0.08, outer => 0.25, extension => 0.25,
            path => [], openings => []};

        my @chain = _chain ($graph_envelope, $source_vertex);
        $molior->{path} = _unpack_chain (@chain);

        for my $edge_id (0 .. scalar @{$molior->{path}} -2)
        {
            _insert_opening ($dom, $molior, $lookup, [@chain], $edge_id);
        }

        YAML::DumpFile ("molior-$yaml_id.yml", $molior);
        $yaml_id++;

        unless (scalar $dom->Levels_Above)
        {
	    my $molior_extrusion = {closed => 0, type => 'molior-extrusion', extension => 0.25,
               elevation => $elevation, height => $height, path => _unpack_chain (@chain),
               cap => [[0,1,2,3], [0,3,4,5], [5,6,0,0], [6,7,8,0]],
               profile => [[0.25,0.0], [0.25,-0.45], [0.3,-0.45], [0.3,-0.4], [0.35,-0.4],
                           [0.35,-0.35], [0.5,-0.25], [0.55,-0.25], [0.55,-0.2]]};

            YAML::DumpFile ("molior-$yaml_id.yml", $molior_extrusion);
            $yaml_id++;
        }
    }

    # do parapets
    for my $source_vertex ($graph_parapet->source_vertices)
    {
        my @chain = _chain ($graph_parapet, $source_vertex);
	my $molior_extrusion = {closed => 0, type => 'molior-extrusion',
           elevation => $elevation, height => 0.0, path => _unpack_chain (@chain),
           cap => [[12,0,1,11], [11,1,2,3], [11,3,4,11], [9,10,11,4], [7,8,9,6], [9,4,6,9], [4,5,6,6]],
           profile => [[0.25,0.0], [0.25,0.71], [0.28,0.71], [0.28,0.74], [0.38,0.875], [0.41,0.875],
                       [0.41,0.9], [-0.14,0.9], [-0.14,0.875], [-0.11,0.875], [-0.11,0.81], [-0.08,0.81], [-0.08,0.0]]};

        YAML::DumpFile ("molior-$yaml_id.yml", $molior_extrusion);
        $yaml_id++;
    }

    # do non-loop party walls
    for my $source_vertex ($graph_partywall->source_vertices)
    {
        my $molior = {closed => 0, captop => 1, capbottom => 1, type => 'molior-wall',
            elevation => $elevation, height => $height, inner => 0.08, outer => 0.25,
            path => [], openings => []};

        $molior->{path} = _unpack_chain (_chain ($graph_partywall, $source_vertex));
        YAML::DumpFile ("molior-$yaml_id.yml", $molior);
        $yaml_id++;
    }

    # do internal walls
    for my $edge ($graph_internal->edges)
    {
        my $molior = {closed => 0, captop => 1, capbottom => 1, type => 'molior-wall',
            elevation => $elevation, height => $height, inner => 0.08, outer => 0.08,
            path => [], openings => []};
        $molior->{path} = _unpack_chain (@{$edge});

        my @types = sort (map {$dom->By_Id ($_)->Type} @{$lookup->{join '-', @$edge}});
        push @{$molior->{openings}}, [{type => 'door_inside', along => '0.18', up => '0.02'}]
            if ($types[0] =~ /^c/i or $types[1] =~ /^c/i or ($types[0] =~ /^k/i and $types[1] =~ /^l/i));

        YAML::DumpFile ("molior-$yaml_id.yml", $molior);
        $yaml_id++;
    }

    # do non-loop covered outside beam
    for my $source_vertex ($graph_covered->source_vertices)
    {
        my @chain = _chain ($graph_covered, $source_vertex);
	my $molior = {closed => 0, type => 'molior-extrusion', extension => -0.08,
           elevation => $elevation, height => 3.0, path => _unpack_chain (@chain),
           cap => [[0,1,10,11], [1,2,3,3], [3,4,5,5], [6,7,8,8], [8,9,10,10], [1,3,8,10], [3,5,6,8]],
           profile => [[0.26,-0.3], [0.26,-0.09], [0.28,-0.09], [0.28,-0.03], [0.31,-0.03], [0.31,0.0],
                       [-0.31,0.0], [-0.31,-0.23], [-0.28,-0.23], [-0.28,-0.27], [-0.26,-0.27], [-0.26,-0.3]]};
        YAML::DumpFile ("molior-$yaml_id.yml", $molior);
        $yaml_id++;
    }
    
    # do columns supporting outside beam
    for my $source_vertex ($graph_covered->source_vertices)
    {
        my $chain = _unpack_chain (_chain ($graph_covered, $source_vertex));
        # corner columns
        for my $corner_id (1 .. scalar @{$chain} -2)
        {
            my $rotation = rad2deg (Urb::Polygon::_angle_2d ($chain->[$corner_id], $chain->[$corner_id +1]));
            my $molior = {type => 'molior-insert', name => 'column_28', scale => [1,1,1],
               elevation => $elevation, height => 0.0, location => $chain->[$corner_id], rotation => $rotation};
            YAML::DumpFile ("molior-$yaml_id.yml", $molior);
            $yaml_id++;
        }
        # intermediate columns
        for my $corner_id (0 .. scalar @{$chain} -2)
        {
            my $rotation = rad2deg (Urb::Polygon::_angle_2d ($chain->[$corner_id], $chain->[$corner_id +1]));
            my $distance = Urb::Polygon::_distance_2d ($chain->[$corner_id], $chain->[$corner_id +1]);
            my $segments = int ($distance / 2);
            next if $segments < 2;
            for my $segment_id (1 .. $segments -1)
            {
                my $inc = $segment_id / $segments;
                my $location = Urb::Polygon::_add_2d (Urb::Polygon::_scale_2d ($inc, $chain->[$corner_id]), Urb::Polygon::_scale_2d (1-$inc, $chain->[$corner_id +1]));
                my $molior = {type => 'molior-insert', name => 'column_28', scale => [1,1,1],
                   elevation => $elevation, height => 0.0, location => $location, rotation => $rotation};
                YAML::DumpFile ("molior-$yaml_id.yml", $molior);
                $yaml_id++;
            }
        }
    }

    # do non-loop covered party fence
    for my $source_vertex ($graph_private_covered->source_vertices)
    {
        my $molior = {closed => 0, captop => 1, capbottom => 1, type => 'molior-wall',
            elevation => $elevation, height => 3.0, inner => 0.08, outer => 0.25, extension => 0.25,
            path => [], openings => []};

        $molior->{path} = _unpack_chain (_chain ($graph_private_covered, $source_vertex));
        YAML::DumpFile ("molior-$yaml_id.yml", $molior);
        $yaml_id++;
    }

    # level 0 stuff only after this
    next if scalar $dom->Levels_Below;

    # do non-loop party fence
    for my $source_vertex ($graph_private->source_vertices)
    {
        my $molior = {closed => 0, captop => 1, capbottom => 1, type => 'molior-wall',
            elevation => $elevation, height => 1.8, inner => -0.15, outer => 0.25, extension => 0.25,
            path => [], openings => []};

        $molior->{path} = _unpack_chain (_chain ($graph_private, $source_vertex));
        YAML::DumpFile ("molior-$yaml_id.yml", $molior);
        $yaml_id++;
    }
}

exit 0;

sub _insert_opening
{
    my ($dom, $molior, $lookup, $chain, $edge_id) = @_;

    # size the list so we can access $molior->{openings}->[-1];
    $molior->{openings}->[scalar @{$chain} -1] = [];
    my $length = _distance_2d ($molior->{path}->[$edge_id], $molior->{path}->[$edge_id +1]);
    my $edge = [$chain->[$edge_id], $chain->[$edge_id +1]];
    my $ids = $lookup->{join '-', @$edge};
    if (grep /^[abcd]$/, @{$ids})
    {
        if ($length < 2)
        {
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_small', along => '0.5', up => '1.0'};
        }
        elsif ($length < 3)
        {
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_narrow', along => '0.5', up => '0.6'};
        }
        else
        {
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_narrow', along => '0.5', up => '0.6'};
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_big', along => '0.5', up => '0.6'};
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_narrow', along => '0.5', up => '0.6'};
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_big', along => '0.5', up => '0.6'};
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_narrow', along => '0.5', up => '0.6'};
        }
    }
    else
    {
        my $type = $dom->By_Id ($ids->[0])->Type;
        if ($type =~ /^[klc]/i and $dom->By_Id ($ids->[1])->Public_Access)
        {
            push @{$molior->{openings}->[$edge_id]}, {type => 'door_panelled', along => '0.5', up => '0.02'};
        }
        elsif (!$dom->Level and $type =~ /^c/i and !$dom->By_Id ($ids->[1])->Public_Access)
        {
            push @{$molior->{openings}->[$edge_id]}, {type => 'door_panelled', along => '0.5', up => '0.02'};
        }
        elsif (!$dom->Level and $type =~ /^[klb]/i and !$dom->By_Id ($ids->[1])->Public_Access)
        {
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_narrow', along => '0.5', up => '0.6'};
            push @{$molior->{openings}->[$edge_id]}, {type => 'door_panelled', along => '0.5', up => '0.02'};
        }
        elsif ($dom->Level and $type =~ /^[klcb]/i and $dom->By_Id ($ids->[1])->Is_Supported)
        {
            push @{$molior->{openings}->[$edge_id]}, {type => 'door_panelled', along => '0.5', up => '0.02'};
        }
        if ($length < 3.5)
        {
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_narrow', along => '0.5', up => '0.6'};
        }
        elsif ($length < 4.5)
        {
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_big', along => '0.5', up => '0.6'};
        }
        else
        {
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_narrow', along => '0.5', up => '0.6'};
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_big', along => '0.5', up => '0.6'};
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_narrow', along => '0.5', up => '0.6'};
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_big', along => '0.5', up => '0.6'};
            push @{$molior->{openings}->[$edge_id]}, {type => 'sash_narrow', along => '0.5', up => '0.6'};
        }
    }

}

sub _unpack_chain
{
    my @chain = @_;
    my @unpacked;
    for my $id (0 .. scalar @chain -1)
    {
         my $packed = $chain[$id];
         my $coor = [split ':', $packed];
         push @unpacked, $coor;
    }
    return [@unpacked];
}

sub _chain
{
    my $g = shift;
    my @list = @_;
    my ($successor) = $g->successors ($list[-1]);
    @list = _chain ($g, @list, $successor) if defined $successor;
    return @list;
}

sub _distance_2d
{
    my ($a, $b) = @_;
    sqrt (($a->[0] - $b->[0])**2 + ($a->[1] - $b->[1])**2);
}

