#!/usr/bin/perl

eval 'exec /usr/bin/perl  -S $0 ${1+"$@"}'
    if 0; # not running under some shell

use strict;
use warnings;
use lib qw/lib/;
use Urb::Dom;
use Urb::Dom::Fitness;
use Urb::Dom::Mutate;
use YAML;

$SIG{INT} = \&interrupt;

my $max_iterations = $ENV{MAX_ITERATIONS} || 1000;

my $path_yaml = shift || die "Mutates an individual and keeps any mutations that increase fitness.
Useful environment variables:
  MAX_ITERATIONS=1000    varies the number of iterations
  CHECKPOINT=1           keeps a copy of all intermediate steps
  START_ITERATION=1001   Useful for continuing after checkpointing
Interrupting the tool writes the results to disk before exiting
Usage: $0 start.yml [mutated.yml]";
my $path_yaml_out = shift || 'temp.yml';
$path_yaml_out =~ s/\.yml$//i;

my $dom = new Urb::Dom;
$dom->Deserialise (YAML::LoadFile ($path_yaml));

my $assessor = new Urb::Dom::Fitness;
my $mutator = new Urb::Dom::Mutate;

my $rating = sprintf ("%.40f", $assessor->_apply ($dom));
my $rating_original = $rating;
print "original: $rating_original\n";
my $improvements = 0;
my $iterations_unproductive = 0;
my $start_iteration = $ENV{START_ITERATION} || 0;

for my $iteration ($start_iteration .. $max_iterations)
{
    my $dom_new = $mutator->apply_rearrange ($dom);
    $iterations_unproductive++;
    for (1 .. int (log ($iterations_unproductive + 20) / log (10)))
    {
        $dom_new = $mutator->apply_rearrange ($dom_new);
    }
    my $rating_new = sprintf ("%.40f", $assessor->_apply ($dom_new));
    print "iteration: $iteration/$max_iterations ($iterations_unproductive): $rating_new\n";
    next unless ($rating_new > $rating);
    $improvements++;
    print "improvements: $improvements; $rating_original -> $rating -> $rating_new\n";
    $rating = $rating_new;
    $dom = $dom_new;
    $iterations_unproductive = 0;
    finish ($path_yaml_out .'_'. sprintf("%08d", $iteration)) if $ENV{CHECKPOINT};
}

finish ($path_yaml_out);

sub interrupt
{
    print STDERR "SIG_INT\n";
    finish ($path_yaml_out);
    exit 1;
}

sub finish
{
    my $path = shift;
    if ($path eq 'temp')
    {
        $path = $dom->Hash;
    }
    print "improvements: $improvements; $rating_original -> $rating\n";
    YAML::DumpFile ($path .'.yml', $dom->Serialise);
    unlink "$path.yml.score";
    open SCORE, ">$path.yml.score";
    $rating = sprintf ("%.40f", $rating);
    print SCORE "$rating\n";
    close SCORE;

    # get list of failures generated during scoring, write to file
    my @failures = $dom->Failures;

    unlink "$path.yml.fails";
    open FAILS, ">$path.yml.fails";
    print FAILS join "\n", @failures, '';
    close FAILS;
}

exit 0;
